"""Globi - Data and simulations for the GloBI project.

A Python package for data science and geospatial analysis.
"""

__version__ = "0.1.0"

# Import main classes and functions (lazy imports to avoid dependency issues)
# Users can import directly from submodules if needed:
# from globi.globi_builder import GloBIBuildingSpec, run_simulation
# from globi.globi_builder import post_process_results, save_results_json
# from globi.experiment_config import GloBISpecInputScythe, simulate_globi_building

__all__ = [
    "__version__",
]
